#define PETSCKSP_DLL

/*
    This file implements the deflated GMRES  
    References:  Erhel, Burrage and Pohl,  Restarted GMRES preconditioned by deflation,J. Computational and Applied Mathematics, 69(1996), 303-318.


 */

#include "dgmresimpl.h"       /*I  "petscksp.h"  I*/
#define GMRES_DELTA_DIRECTIONS 10
#define GMRES_DEFAULT_MAXK     30
//#define PetscTruth PetscBool
static PetscErrorCode    DGMRESGetNewVectors(KSP,PetscInt);
static PetscErrorCode    DGMRESUpdateHessenberg(KSP,PetscInt,PetscTruth,PetscReal*);
static PetscErrorCode    BuildDGmresSoln(PetscScalar*,Vec,Vec,KSP,PetscInt);
/* Functions from gmres */
EXTERN_C_BEGIN
EXTERN PetscErrorCode PETSCKSP_DLLEXPORT KSPComputeExtremeSingularValues_GMRES(KSP,PetscReal *,PetscReal *);
EXTERN PetscErrorCode PETSCKSP_DLLEXPORT KSPComputeEigenvalues_GMRES(KSP,PetscInt,PetscReal *,PetscReal *,PetscInt *);
EXTERN PetscErrorCode PETSCKSP_DLLEXPORT KSPGMRESSetHapTol_GMRES(KSP,double);
EXTERN PetscErrorCode PETSCKSP_DLLEXPORT KSPGMRESSetPreAllocateVectors_GMRES(KSP);
EXTERN PetscErrorCode PETSCKSP_DLLEXPORT KSPGMRESSetRestart_GMRES(KSP,PetscInt);
EXTERN PetscErrorCode PETSCKSP_DLLEXPORT KSPGMRESGetRestart_GMRES(KSP,PetscInt*);
EXTERN PetscErrorCode PETSCKSP_DLLEXPORT KSPGMRESSetOrthogonalization_GMRES(KSP,PetscErrorCode (*)(KSP,PetscInt));
EXTERN PetscErrorCode PETSCKSP_DLLEXPORT KSPGMRESGetOrthogonalization_GMRES(KSP,PetscErrorCode (**)(KSP,PetscInt));
EXTERN PetscErrorCode PETSCKSP_DLLEXPORT KSPGMRESSetCGSRefinementType_GMRES(KSP,KSPGMRESCGSRefinementType);
EXTERN PetscErrorCode PETSCKSP_DLLEXPORT KSPGMRESGetCGSRefinementType_GMRES(KSP,KSPGMRESCGSRefinementType*);
EXTERN PetscErrorCode PETSCKSP_DLLEXPORT KSPSetUp_GMRES(KSP);
EXTERN PetscErrorCode PETSCKSP_DLLEXPORT KSPView_GMRES(KSP,PetscViewer);
EXTERN PetscErrorCode PETSCKSP_DLLEXPORT KSPDestroy_GMRES(KSP);
EXTERN PetscErrorCode PETSCKSP_DLLEXPORT KSPSetFromOptions_GMRES(KSP);
EXTERN_C_END



#undef __FUNCT__
#define __FUNCT__  "KSPDGMRESSetEigen"
PetscErrorCode PETSCKSP_DLLEXPORT KSPDGMRESSetEigen(KSP ksp,PetscInt nb_eig)
{
	PetscErrorCode ierr;

	PetscFunctionBegin;
	ierr = PetscTryMethod((ksp),"KSPDGMRESSetEigen_C",(KSP,PetscInt),(ksp,nb_eig));CHKERRQ(ierr);
	PetscFunctionReturn(0);
}
#undef __FUNCT__
#define __FUNCT__  "KSPDGMRESSetMaxEigen"
PetscErrorCode PETSCKSP_DLLEXPORT KSPDGMRESSetMaxEigen(KSP ksp,PetscInt max_neig)
{
	PetscErrorCode ierr;

	PetscFunctionBegin;
	ierr = PetscTryMethod((ksp),"KSPDGMRESSetMaxEigen_C",(KSP,PetscInt),(ksp,max_neig));CHKERRQ(ierr);
	PetscFunctionReturn(0);
}
#undef __FUNCT__
#define __FUNCT__  "KSPDGMRESComputeSchurForm"
PetscErrorCode PETSCKSP_DLLEXPORT KSPDGMRESComputeSchurForm(KSP ksp,PetscInt *neig)
{
	PetscErrorCode ierr;

	PetscFunctionBegin;
	ierr = PetscTryMethod((ksp),"KSPDGMRESComputeSchurForm_C",(KSP, PetscInt*),(ksp, neig));CHKERRQ(ierr);
	PetscFunctionReturn(0);
}
#undef __FUNCT__
#define __FUNCT__  "KSPDGMRESComputeDeflationData"
PetscErrorCode PETSCKSP_DLLEXPORT KSPDGMRESComputeDeflationData(KSP ksp)
{
	PetscErrorCode ierr;

	PetscFunctionBegin;
	ierr = PetscTryMethod((ksp),"KSPDGMRESComputeDeflationData_C",(KSP),(ksp));CHKERRQ(ierr);
	PetscFunctionReturn(0);
}
#undef __FUNCT__
#define __FUNCT__  "KSPDGMRESApplyDeflation"
PetscErrorCode PETSCKSP_DLLEXPORT KSPDGMRESApplyDeflation(KSP ksp, Vec x, Vec y)
{
	PetscErrorCode ierr;

	PetscFunctionBegin;
	ierr = PetscTryMethod((ksp),"KSPDGMRESApplyDeflation_C",(KSP, Vec, Vec),(ksp, x, y));CHKERRQ(ierr);
	PetscFunctionReturn(0);
}
#undef __FUNCT__
#define __FUNCT__ "KSPDGMRESImproveEig"
PetscErrorCode PETSCKSP_DLLEXPORT KSPDGMRESImproveEig(KSP ksp, PetscInt neig)
{
	PetscErrorCode ierr;
	
	PetscFunctionBegin;
	ierr = PetscTryMethod((ksp), "KSPDGMRESImproveEig_C", (KSP, PetscInt), (ksp, neig)); CHKERRQ(ierr);
	PetscFunctionReturn(0); 
}
#undef __FUNCT__
#define __FUNCT__ "KSPSetUp_DGMRES"
PetscErrorCode    KSPSetUp_DGMRES(KSP ksp)
{
  PetscErrorCode    ierr;
  KSP_DGMRES        *dgmres 	= (KSP_DGMRES *)ksp->data;
  PetscInt			neig		= dgmres->neig+EIG_OFFSET;
  PetscInt			max_k		= dgmres->max_k+1; 

  PetscFunctionBegin;
  
  ierr          = KSPSetUp_GMRES(ksp);CHKERRQ(ierr);
  if(dgmres->neig==0) 
	  PetscFunctionReturn(0);
  
  /* Allocate workspace for the Schur vectors*/
  ierr=PetscMalloc((neig)*max_k*sizeof(PetscReal), &SR); CHKERRQ(ierr);
  UU = PETSC_NULL;  	XX = PETSC_NULL; 	MX = PETSC_NULL; AUU = PETSC_NULL;
  XMX = PETSC_NULL; 	XMU = PETSC_NULL; 	UMX = PETSC_NULL; AUAU = PETSC_NULL;
  TT = PETSC_NULL; 		TTF = PETSC_NULL; 	INVP = PETSC_NULL; 
  X1 = PETSC_NULL; 		X2 = PETSC_NULL; 	MU = PETSC_NULL;
  PetscFunctionReturn(0);
}

/*
    Run GMRES, possibly with restart.  Return residual history if requested.
    input parameters:

.	gmres  - structure containing parameters and work areas

    output parameters:
.        nres    - residuals (from preconditioned system) at each step.
                  If restarting, consider passing nres+it.  If null, 
                  ignored
.        itcount - number of iterations used.  nres[0] to nres[itcount]
                  are defined.  If null, ignored.
		  
    Notes:
    On entry, the value in vector VEC_VV(0) should be the initial residual
    (this allows shortcuts where the initial preconditioned residual is 0).
 */
#undef __FUNCT__  
#define __FUNCT__ "DGMREScycle"
PetscErrorCode DGMREScycle(PetscInt *itcount,KSP ksp)
{
  KSP_DGMRES      *dgmres = (KSP_DGMRES *)(ksp->data);
  PetscReal      res_norm,res,hapbnd,tt;
  PetscErrorCode ierr;
  PetscInt       it = 0, max_k = dgmres->max_k;
  PetscTruth     hapend = PETSC_FALSE;
  PetscReal		res_old;
  PetscInt		test;

  PetscFunctionBegin;
  ierr    = VecNormalize(VEC_VV(0),&res_norm);CHKERRQ(ierr);
  res     = res_norm;
  *GRS(0) = res_norm;

  /* check for the convergence */
  ierr = PetscObjectTakeAccess(ksp);CHKERRQ(ierr);
  ksp->rnorm = res;
  ierr = PetscObjectGrantAccess(ksp);CHKERRQ(ierr);
  dgmres->it = (it - 1);
  KSPLogResidualHistory(ksp,res);
  KSPMonitor(ksp,ksp->its,res); 
  if (!res) {
    if (itcount) *itcount = 0;
    ksp->reason = KSP_CONVERGED_ATOL;
    ierr = PetscInfo(ksp,"Converged due to zero residual norm on entry\n");CHKERRQ(ierr);
    PetscFunctionReturn(0);
  }
//   /* record the residual norm to test if deflation is needed */
	res_old = res;
  
  ierr = (*ksp->converged)(ksp,ksp->its,res,&ksp->reason,ksp->cnvP);CHKERRQ(ierr);
  while (!ksp->reason && it < max_k && ksp->its < ksp->max_it) {
    if (it) {
      KSPLogResidualHistory(ksp,res);
      KSPMonitor(ksp,ksp->its,res); 
    }
    dgmres->it = (it - 1);
    if (dgmres->vv_allocated <= it + VEC_OFFSET + 1) {
      ierr = DGMRESGetNewVectors(ksp,it+1);CHKERRQ(ierr);
    }
	if( dgmres->r > 0 ) { 
		if(ksp->pc_side == PC_LEFT){
			/* Apply the first preconditioner */
			ierr = KSP_PCApplyBAorAB(ksp,VEC_VV(it), VEC_TEMP,VEC_TEMP_MATOP);CHKERRQ(ierr);
			/* Then apply Deflation as a preconditioner */
			ierr=KSPDGMRESApplyDeflation(ksp, VEC_TEMP, VEC_VV(1+it)); CHKERRQ(ierr);
		}
		else if(ksp->pc_side == PC_RIGHT){
			ierr=KSPDGMRESApplyDeflation(ksp, VEC_VV(it), VEC_TEMP); CHKERRQ(ierr);
			ierr=KSP_PCApplyBAorAB(ksp, VEC_TEMP, VEC_VV(1+it), VEC_TEMP_MATOP); CHKERRQ(ierr);
		}
	}
	else{
		ierr = KSP_PCApplyBAorAB(ksp,VEC_VV(it),VEC_VV(1+it),VEC_TEMP_MATOP);CHKERRQ(ierr);
	}
	dgmres->matvecs += 1;
    /* update hessenberg matrix and do Gram-Schmidt */
    ierr = (*dgmres->orthog)(ksp,it);CHKERRQ(ierr);

    /* vv(i+1) . vv(i+1) */
    ierr = VecNormalize(VEC_VV(it+1),&tt);CHKERRQ(ierr);
    /* save the magnitude */
    *HH(it+1,it)    = tt;
    *HES(it+1,it)   = tt;

    /* check for the happy breakdown */
    hapbnd  = PetscAbsScalar(tt / *GRS(it));
    if (hapbnd > dgmres->haptol) hapbnd = dgmres->haptol;
    if (tt < hapbnd) {
      ierr = PetscInfo2(ksp,"Detected happy breakdown, current hapbnd = %G tt = %G\n",hapbnd,tt);CHKERRQ(ierr);
      hapend = PETSC_TRUE;
    }
    ierr = DGMRESUpdateHessenberg(ksp,it,hapend,&res);CHKERRQ(ierr);

    it++;
    dgmres->it  = (it-1);  /* For converged */
    ksp->its++;
    ksp->rnorm = res;
    if (ksp->reason) break;

    ierr = (*ksp->converged)(ksp,ksp->its,res,&ksp->reason,ksp->cnvP);CHKERRQ(ierr);

    /* Catch error in happy breakdown and signal convergence and break from loop */
    if (hapend) {
      if (!ksp->reason) {
        SETERRQ1(PETSC_ERR_PLIB,"You reached the happy break down, but convergence was not indicated. Residual norm = %G",res);
      }
      break;
    }
  }

  /* Monitor if we know that we will not return for a restart */
  if (it && (ksp->reason || ksp->its >= ksp->max_it)) {
    KSPLogResidualHistory(ksp,res);
    KSPMonitor(ksp,ksp->its,res);
  }
 
  if (itcount) *itcount    = it;


  /*
    Down here we have to solve for the "best" coefficients of the Krylov
    columns, add the solution values together, and possibly unwind the
    preconditioning from the solution
   */
  /* Form the solution (or the solution so far) */
  ierr = BuildDGmresSoln(GRS(0),ksp->vec_sol,ksp->vec_sol,ksp,it-1);CHKERRQ(ierr);
  
  /* Compute data for the deflation to be used during the next restart */
  if(!ksp->reason && ksp->its < ksp->max_it) { 
	  test = max_k *log(ksp->rtol/res)/log(res/res_old);
	  /* Compute data for the deflation if the residual rtol will not be reached in the remaining number of steps allowed  */
	  if( (test > dgmres->smv*(ksp->max_it-ksp->its)) || dgmres->force){
		  ierr= KSPDGMRESComputeDeflationData(ksp); CHKERRQ(ierr);
	  }
  }

  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "KSPSolve_DGMRES"
PetscErrorCode KSPSolve_DGMRES(KSP ksp)
{
  PetscErrorCode ierr;
  PetscInt       its,itcount;
  KSP_DGMRES      *dgmres = (KSP_DGMRES *)ksp->data;
  PetscTruth     guess_zero = ksp->guess_zero;

  PetscFunctionBegin;
  if (ksp->calc_sings && !dgmres->Rsvd) SETERRQ(PETSC_ERR_ORDER,"Must call KSPSetComputeSingularValues() before KSPSetUp() is called");
  if (ksp->normtype != KSP_NORM_PRECONDITIONED && ksp->pc_side != PC_RIGHT) SETERRQ(PETSC_ERR_ARG_WRONGSTATE,"Use right preconditioning -ksp_pc_side right if want unpreconditioned norm)");

  ierr     = PetscObjectTakeAccess(ksp);CHKERRQ(ierr);
  ksp->its = 0;
  dgmres->matvecs = 0;
  ierr     = PetscObjectGrantAccess(ksp);CHKERRQ(ierr);

  itcount     = 0;
  ksp->reason = KSP_CONVERGED_ITERATING;
  while (!ksp->reason) {
    ierr     = KSPInitialResidual(ksp,ksp->vec_sol,VEC_TEMP,VEC_TEMP_MATOP,VEC_VV(0),ksp->vec_rhs);CHKERRQ(ierr);
	if (ksp->pc_side == PC_LEFT){
		dgmres->matvecs += 1;
		if(dgmres->r > 0 ){
		  ierr = KSPDGMRESApplyDeflation(ksp, VEC_VV(0), VEC_TEMP); CHKERRQ(ierr);
		  ierr=VecCopy(VEC_TEMP, VEC_VV(0)); CHKERRQ(ierr);
		}
	}
	
    ierr     = DGMREScycle(&its,ksp);CHKERRQ(ierr);
    itcount += its;  
    if (itcount >= ksp->max_it) {
      if (!ksp->reason) ksp->reason = KSP_DIVERGED_ITS;
      break;
    }
    ksp->guess_zero = PETSC_FALSE; /* every future call to KSPInitialResidual() will have nonzero guess */
  }
  ksp->guess_zero = guess_zero; /* restore if user provided nonzero initial guess */
  PetscFunctionReturn(0);
}


#undef __FUNCT__  
#define __FUNCT__ "KSPDestroy_DGMRES" 
PetscErrorCode KSPDestroy_DGMRES(KSP ksp)
{  
  PetscErrorCode ierr;
  KSP_DGMRES     *dgmres = (KSP_DGMRES*)ksp->data;
  PetscInt 		neig1 = dgmres->neig+EIG_OFFSET;
  PetscInt		max_neig = dgmres->max_neig;

  PetscFunctionBegin;
  if(dgmres->r){
	  ierr=VecDestroyVecs(UU, max_neig); CHKERRQ(ierr);
	  ierr=VecDestroyVecs(MU, max_neig); CHKERRQ(ierr);
	  ierr=VecDestroyVecs(XX, neig1); CHKERRQ(ierr);
	  ierr=VecDestroyVecs(MX, neig1); CHKERRQ(ierr);
	  
	  ierr=PetscFree(TT); CHKERRQ(ierr);
	  ierr=PetscFree(TTF); CHKERRQ(ierr);
	  ierr=PetscFree(INVP); CHKERRQ(ierr);
	  
	  ierr=PetscFree(SR); CHKERRQ(ierr);
	  ierr=PetscFree(XMX); CHKERRQ(ierr);
	  ierr=PetscFree(UMX); CHKERRQ(ierr);
	  ierr=PetscFree(XMU); CHKERRQ(ierr);
	  ierr=PetscFree(X1); CHKERRQ(ierr);
	  ierr=PetscFree(X2); CHKERRQ(ierr);
	  ierr=PetscFree(dgmres->work); CHKERRQ(ierr);
	  ierr=PetscFree(dgmres->iwork); CHKERRQ(ierr);
	  ierr=PetscFree(ORTH); CHKERRQ(ierr);
	  
	  ierr = PetscFree(AUAU); CHKERRQ(ierr);
	  ierr = PetscFree(AUU); CHKERRQ(ierr);
	  ierr = PetscFree(SR2); CHKERRQ(ierr);
  }
  ierr = KSPDestroy_GMRES(ksp);CHKERRQ(ierr);
  PetscFunctionReturn(0);
}
/*
    BuildDGmresSoln - create the solution from the starting vector and the
    current iterates.

    Input parameters:
        nrs - work area of size it + 1.
	vs  - index of initial guess
	vdest - index of result.  Note that vs may == vdest (replace
	        guess with the solution).

     This is an internal routine that knows about the GMRES internals.
 */
#undef __FUNCT__  
#define __FUNCT__ "BuildDGmresSoln"
static PetscErrorCode BuildDGmresSoln(PetscScalar* nrs,Vec vs,Vec vdest,KSP ksp,PetscInt it)
{
	PetscScalar    tt;
	PetscErrorCode ierr;
	PetscInt       ii,k,j;
	KSP_DGMRES      *dgmres = (KSP_DGMRES *)(ksp->data);

	PetscFunctionBegin;
	/* Solve for solution vector that minimizes the residual */

	/* If it is < 0, no gmres steps have been performed */
	if (it < 0) {
		ierr = VecCopy(vs,vdest);CHKERRQ(ierr); /* VecCopy() is smart, exists immediately if vguess == vdest */
		PetscFunctionReturn(0);
	}
	if (*HH(it,it) == 0.0) SETERRQ2(PETSC_ERR_CONV_FAILED,"Likley your matrix is the zero operator. HH(it,it) is identically zero; it = %D GRS(it) = %G",it,PetscAbsScalar(*GRS(it)));
	if (*HH(it,it) != 0.0) {
		nrs[it] = *GRS(it) / *HH(it,it);
	} else {
		nrs[it] = 0.0;
	}
	for (ii=1; ii<=it; ii++) {
		k   = it - ii;
		tt  = *GRS(k);
		for (j=k+1; j<=it; j++) tt  = tt - *HH(k,j) * nrs[j];
		if (*HH(k,k) == 0.0) SETERRQ2(PETSC_ERR_CONV_FAILED,"Likley your matrix is singular. HH(k,k) is identically zero; it = %D k = %D",it,k);
		nrs[k]   = tt / *HH(k,k);
	}

	/* Accumulate the correction to the solution of the preconditioned problem in TEMP */
	ierr = VecSet(VEC_TEMP,0.0);CHKERRQ(ierr);
	ierr = VecMAXPY(VEC_TEMP,it+1,nrs,&VEC_VV(0));CHKERRQ(ierr);
	
	/* Apply deflation */
	if(ksp->pc_side==PC_RIGHT && dgmres->r > 0){
		ierr=KSPDGMRESApplyDeflation(ksp, VEC_TEMP, VEC_TEMP_MATOP); CHKERRQ(ierr);
		ierr=VecCopy(VEC_TEMP_MATOP, VEC_TEMP); CHKERRQ(ierr);
	}
	ierr = KSPUnwindPreconditioner(ksp,VEC_TEMP,VEC_TEMP_MATOP);CHKERRQ(ierr);
	
	
	/* add solution to previous solution */
	if (vdest != vs) {
		ierr = VecCopy(vs,vdest);CHKERRQ(ierr);
	}
	ierr = VecAXPY(vdest,1.0,VEC_TEMP);CHKERRQ(ierr);
	PetscFunctionReturn(0);
}
/*
   Do the scalar work for the orthogonalization.  Return new residual norm.
 */
#undef __FUNCT__  
#define __FUNCT__ "DGMRESUpdateHessenberg"
static PetscErrorCode DGMRESUpdateHessenberg(KSP ksp,PetscInt it,PetscTruth hapend,PetscReal *res)
{
	PetscScalar *hh,*cc,*ss,tt;
	PetscInt    j;
	KSP_DGMRES   *dgmres = (KSP_DGMRES *)(ksp->data);

	PetscFunctionBegin;
	hh  = HH(0,it);
	cc  = CC(0);
	ss  = SS(0);

  /* Apply all the previously computed plane rotations to the new column
	of the Hessenberg matrix */
	for (j=1; j<=it; j++) {
		tt  = *hh;
#if defined(PETSC_USE_COMPLEX)
		*hh = PetscConj(*cc) * tt + *ss * *(hh+1);
#else
		*hh = *cc * tt + *ss * *(hh+1);
#endif
		hh++;
		*hh = *cc++ * *hh - (*ss++ * tt);
	}

  /*
    compute the new plane rotation, and apply it to:
	1) the right-hand-side of the Hessenberg system
	2) the new column of the Hessenberg matrix
	thus obtaining the updated value of the residual
  */
	if (!hapend) {
#if defined(PETSC_USE_COMPLEX)
		tt        = PetscSqrtScalar(PetscConj(*hh) * *hh + PetscConj(*(hh+1)) * *(hh+1));
#else
		tt        = PetscSqrtScalar(*hh * *hh + *(hh+1) * *(hh+1));
#endif
		if (tt == 0.0) {
			ksp->reason = KSP_DIVERGED_NULL;
			PetscFunctionReturn(0);
		}
		*cc       = *hh / tt;
		*ss       = *(hh+1) / tt;
		*GRS(it+1) = - (*ss * *GRS(it));
#if defined(PETSC_USE_COMPLEX)
		*GRS(it)   = PetscConj(*cc) * *GRS(it);
		*hh       = PetscConj(*cc) * *hh + *ss * *(hh+1);
#else
		*GRS(it)   = *cc * *GRS(it);
		*hh       = *cc * *hh + *ss * *(hh+1);
#endif
		*res      = PetscAbsScalar(*GRS(it+1));
	} else {
    /* happy breakdown: HH(it+1, it) = 0, therfore we don't need to apply 
		another rotation matrix (so RH doesn't change).  The new residual is 
		always the new sine term times the residual from last time (GRS(it)), 
		but now the new sine rotation would be zero...so the residual should
		be zero...so we will multiply "zero" by the last residual.  This might
		not be exactly what we want to do here -could just return "zero". */
 
		*res = 0.0;
	}
	PetscFunctionReturn(0);
}
/*
   This routine allocates more work vectors, starting from VEC_VV(it).
 */
#undef __FUNCT__  
#define __FUNCT__ "DGMRESGetNewVectors" 
static PetscErrorCode DGMRESGetNewVectors(KSP ksp,PetscInt it)
{
	KSP_DGMRES      *dgmres = (KSP_DGMRES *)ksp->data;
	PetscErrorCode ierr;
	PetscInt       nwork = dgmres->nwork_alloc,k,nalloc;

	PetscFunctionBegin;
	nalloc = PetscMin(ksp->max_it,dgmres->delta_allocate);
  /* Adjust the number to allocate to make sure that we don't exceed the
	number of available slots */
	if (it + VEC_OFFSET + nalloc >= dgmres->vecs_allocated){
		nalloc = dgmres->vecs_allocated - it - VEC_OFFSET;
	}
	if (!nalloc) PetscFunctionReturn(0);

	dgmres->vv_allocated += nalloc;
	ierr = KSPGetVecs(ksp,nalloc,&dgmres->user_work[nwork],0,PETSC_NULL);CHKERRQ(ierr);
	ierr = PetscLogObjectParents(ksp,nalloc,dgmres->user_work[nwork]);CHKERRQ(ierr);
	dgmres->mwork_alloc[nwork] = nalloc;
	for (k=0; k<nalloc; k++) {
		dgmres->vecs[it+VEC_OFFSET+k] = dgmres->user_work[nwork][k];
	}
	dgmres->nwork_alloc++;
	PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "KSPBuildSolution_DGMRES"
PetscErrorCode KSPBuildSolution_DGMRES(KSP ksp,Vec  ptr,Vec *result)
{
	KSP_DGMRES      *dgmres = (KSP_DGMRES *)ksp->data; 
	PetscErrorCode ierr;

	PetscFunctionBegin;
	if (!ptr) {
		if (!dgmres->sol_temp) {
			ierr = VecDuplicate(ksp->vec_sol,&dgmres->sol_temp);CHKERRQ(ierr);
			ierr = PetscLogObjectParent(ksp,dgmres->sol_temp);CHKERRQ(ierr);
		}
		ptr = dgmres->sol_temp;
	}
	if (!dgmres->nrs) {
		/* allocate the work area */
		ierr = PetscMalloc(dgmres->max_k*sizeof(PetscScalar),&dgmres->nrs);CHKERRQ(ierr);
		ierr = PetscLogObjectMemory(ksp,dgmres->max_k*sizeof(PetscScalar));CHKERRQ(ierr);
	}

	ierr = BuildDGmresSoln(dgmres->nrs,ksp->vec_sol,ptr,ksp,dgmres->it);CHKERRQ(ierr);
	if (result) *result = ptr;
	PetscFunctionReturn(0);
}


#undef __FUNCT__  
#define __FUNCT__ "KSPView_DGMRES" 
PetscErrorCode KSPView_DGMRES(KSP ksp,PetscViewer viewer)
{
  KSP_DGMRES      *dgmres = (KSP_DGMRES *)ksp->data; 
  PetscErrorCode ierr;
  PetscTruth     iascii;

  PetscFunctionBegin;
  ierr = KSPView_GMRES(ksp,viewer);CHKERRQ(ierr);
  ierr = PetscTypeCompare((PetscObject)viewer,PETSC_VIEWER_ASCII,&iascii);CHKERRQ(ierr);
  if (iascii) {
    if(dgmres->force)
      PetscViewerASCIIPrintf(viewer, "       DGMRES: Adaptive strategy is used: FALSE\n");
    else
      PetscViewerASCIIPrintf(viewer, "       DGMRES: Adaptive strategy is used: TRUE\n");
    ierr=PetscViewerASCIIPrintf(viewer, "	DGMRES: Frequency of extracted eigenvalues = %D\n", dgmres->neig); CHKERRQ(ierr);
    ierr=PetscViewerASCIIPrintf(viewer, "	DGMRES: Total number of extracted eigenvalues = %D\n", dgmres->r); CHKERRQ(ierr);
    ierr=PetscViewerASCIIPrintf(viewer, "	DGMRES: Maximum number of eigenvalues set to be extracted = %D\n", dgmres->max_neig); CHKERRQ(ierr);
    ierr=PetscViewerASCIIPrintf(viewer, "	DGMRES: relaxation parameter for the adaptive strategy(smv)  = %g\n", dgmres->smv); CHKERRQ(ierr);
    ierr=PetscViewerASCIIPrintf(viewer, "	DGMRES: Number of matvecs : %D\n", dgmres->matvecs); CHKERRQ(ierr);
  } else {
    SETERRQ1(PETSC_ERR_SUP,"Viewer type %s not supported for KSP DGMRES",((PetscObject)viewer)->type_name);
  }

  PetscFunctionReturn(0);
}

/* New DGMRES functions */

EXTERN_C_BEGIN
#undef __FUNCT__  
#define __FUNCT__ "KSPDGMRESSetEigen_DGMRES" 
PetscErrorCode PETSCKSP_DLLEXPORT KSPDGMRESSetEigen_DGMRES(KSP ksp,PetscInt neig)
{
	KSP_DGMRES      *dgmres = (KSP_DGMRES *)ksp->data;
	PetscFunctionBegin;
	if ( neig< 0 && neig >dgmres->max_k)  SETERRQ(PETSC_ERR_ARG_OUTOFRANGE,"The value of neig must be positive and less than the restart value ");
	dgmres->neig=neig;
	PetscFunctionReturn(0);
}
EXTERN_C_END
	
EXTERN_C_BEGIN
#undef __FUNCT__  
#define __FUNCT__ "KSPDGMRESSetMaxEigen_DGMRES" 
PetscErrorCode PETSCKSP_DLLEXPORT KSPDGMRESSetMaxEigen_DGMRES(KSP ksp,PetscInt max_neig)
{
	KSP_DGMRES      *dgmres = (KSP_DGMRES *)ksp->data;
	PetscFunctionBegin;
	if ( max_neig < 0 && max_neig >dgmres->max_k)  SETERRQ(PETSC_ERR_ARG_OUTOFRANGE,"The value of max_neig must be positive and less than the restart value ");
	dgmres->max_neig=max_neig;
	PetscFunctionReturn(0);
}
EXTERN_C_END

extern PetscErrorCode KSPSetFromOptions_DGMRES(KSP);

#undef __FUNCT__  
#define __FUNCT__ "KSPSetFromOptions_DGMRES"
PetscErrorCode KSPSetFromOptions_DGMRES(KSP ksp)
{
  PetscErrorCode ierr;
  PetscInt       neig;
  PetscInt	     max_neig;
  KSP_DGMRES     *dgmres = (KSP_DGMRES*)ksp->data;
  PetscTruth     flg;
  PetscReal		smv;
  PetscInt		input;

  PetscFunctionBegin;
  ierr = KSPSetFromOptions_GMRES(ksp);CHKERRQ(ierr);
  ierr = PetscOptionsHead("KSP DGMRES Options");CHKERRQ(ierr);
    ierr = PetscOptionsInt("-ksp_dgmres_eigen","Number of smallest eigenvalues to extract at each restart","KSPDGMRESSetEigen",dgmres->neig, &neig, &flg);CHKERRQ(ierr);
	if (flg) { ierr = KSPDGMRESSetEigen(ksp, neig);CHKERRQ(ierr); }
	ierr = PetscOptionsInt("-ksp_dgmres_max_eigen","Maximum Number of smallest eigenvalues to extract ","KSPDGMRESSetMaxEigen",dgmres->max_neig, &max_neig, &flg);CHKERRQ(ierr);
	if (flg) {ierr = KSPDGMRESSetMaxEigen(ksp, max_neig);CHKERRQ(ierr); }
 	ierr=PetscOptionsReal("-ksp_dgmres_ratio", "Smaller number of matrix-vectors product", NULL, dgmres->smv, &smv, &flg); CHKERRQ(ierr);
	if(flg) dgmres->smv = smv;
	ierr = PetscOptionsInt("-ksp_dgmres_improve", "Improve the computation of eigenvalues by solving a new generalized eigenvalue problem", NULL, dgmres->improve, &input, &flg); CHKERRQ(ierr);
	if(flg) dgmres->improve = input; 
	ierr = PetscOptionsInt("-ksp_dgmres_force", "Sets DGMRES active, i.e do not use the adaptive strategy", NULL, dgmres->force, &input, &flg); CHKERRQ(ierr);
	if(flg) dgmres->force = input; 
  ierr = PetscOptionsTail();CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

EXTERN_C_BEGIN
#undef __FUNCT__
#define __FUNCT__ "KSPDGMRESComputeDeflationData_DGMRES"
PetscErrorCode PETSCKSP_DLLEXPORT KSPDGMRESComputeDeflationData_DGMRES(KSP ksp)
{
	KSP_DGMRES     	*dgmres = (KSP_DGMRES*)ksp->data;
	PetscErrorCode 	ierr; 
	PetscInt	i,j, k;
	PetscInt	r=dgmres->r;
	PetscBLASInt	br, bmax, info; 
	PetscInt	neig; /* number of eigenvalues to extract at each restart */
	PetscInt	neig1 = dgmres->neig + EIG_OFFSET; /* max number of eig that can be extracted at each restart */
	PetscInt	max_neig = dgmres->max_neig; /* Max number of eigenvalues to extract during the iterative process */
	PetscInt	N = dgmres->max_k+1;
	PetscInt    n = dgmres->it+1;
	PetscReal	alpha; 
	
	PetscFunctionBegin;
	ierr=PetscLogEventBegin(KSP_DGMRESComputeDeflationData, ksp, 0,0,0); CHKERRQ(ierr);
	if( (dgmres->neig == 0) )
		PetscFunctionReturn(0);
	if( max_neig < (r+neig1) && !dgmres->improve ){
		ierr=PetscLogEventEnd(KSP_DGMRESComputeDeflationData, ksp, 0,0,0); CHKERRQ(ierr);
		PetscFunctionReturn(0);
	}
	
	ierr= KSPDGMRESComputeSchurForm(ksp, &neig); CHKERRQ(ierr);
	
	/* Form the extended Schur vectors X=VV*Sr */
	if(!XX){
		ierr=VecDuplicateVecs(VEC_VV(0), neig1, &XX); CHKERRQ(ierr);
	}
	for (j = 0; j<neig; j++){
		ierr = VecZeroEntries(XX[j]); CHKERRQ(ierr);
		ierr = VecMAXPY(XX[j], n, &SR[j*N], &VEC_VV(0)); CHKERRQ(ierr);
	}
	
	/* Orthogonalize X against U */
	if(!ORTH){
		ierr=PetscMalloc(max_neig*sizeof(PetscReal), &ORTH); CHKERRQ(ierr);
	}
	if ( r > 0){
		/* modified Gram-Schmidt */
		for (j = 0; j<neig; j++) { 
			for (i=0; i<r; i++){
				/* First, compute U'*X[j] */
				ierr = VecDot(XX[j], UU[i], &alpha); CHKERRQ(ierr);
				/* Then, compute X(j)=X(j)-U*U'*X(j) */
				ierr = VecAXPY(XX[j], -alpha, UU[i]); CHKERRQ(ierr);
			}
		}	
	}
	/* Compute MX = M^{-1}*A*X */
	if(!MX){
		ierr=VecDuplicateVecs(VEC_VV(0), neig1, &MX); CHKERRQ(ierr);
	}
	for (j = 0; j<neig; j++){
		ierr = KSP_PCApplyBAorAB(ksp, XX[j], MX[j], VEC_TEMP_MATOP); CHKERRQ(ierr);
	}
	dgmres->matvecs += neig;
	
	if( (r+neig1) > max_neig && dgmres->improve){ /* Improve the approximate eigenvectors in X by solving a new generalized eigenvalue -- Quite expensive to do this actually */
		ierr = KSPDGMRESImproveEig(ksp, neig); CHKERRQ(ierr);
		PetscFunctionReturn(0); /* We return here since data for M have been improved in  KSPDGMRESImproveEig()*/
	}
	
	/* Compute XMX = X'*M^{-1}*A*X -- size (neig, neig) */
	if(!XMX){
		ierr=PetscMalloc(neig1*neig1*sizeof(PetscReal), &XMX); CHKERRQ(ierr);
	}
	for(j = 0; j < neig; j++){
		ierr=VecMDot(MX[j], neig, XX, &(XMX[j*neig1]) ); CHKERRQ(ierr);
	}
	
	if( r > 0 ){
		/* Compute UMX = U'*M^{-1}*A*X -- size (r, neig) */
		if(!UMX){
			ierr = PetscMalloc(max_neig*neig1*sizeof(PetscReal), &UMX); CHKERRQ(ierr);
		}
		for(j = 0; j < neig; j++){
			ierr = VecMDot(MX[j], r, UU, &(UMX[j*max_neig]) ); CHKERRQ(ierr);
		}
		/* Compute XMU = X'*M^{-1}*A*U -- size (neig, r) */
		if(!XMU){
			ierr = PetscMalloc(max_neig*neig1*sizeof(PetscReal), &XMU); CHKERRQ(ierr);
		}
		for(j = 0; j<r; j++){
			ierr=VecMDot(MU[j], neig, XX, &(XMU[j*neig1]) ); CHKERRQ(ierr);
		}
		
	}
	
	/* Form the new matrix T = [T UMX; XMU XMX]; */
	if(!TT){
		ierr = PetscMalloc(max_neig*max_neig*sizeof(PetscReal), &TT); CHKERRQ(ierr);
	}
	if( r > 0){
		/* Add XMU to T */
		for(j = 0; j < r; j++){
			ierr = PetscMemcpy(&(TT[max_neig*j+r]), &(XMU[neig1*j]), neig*sizeof(PetscReal));
			CHKERRQ(ierr);
		}
		/* Add [UMX; XMX] to T */
		for(j = 0; j < neig; j++){
			k = r+j;
			ierr = PetscMemcpy(&(TT[max_neig*k]), &(UMX[max_neig*j]), r*sizeof(PetscReal));
			CHKERRQ(ierr);
			ierr=PetscMemcpy(&(TT[max_neig*k + r]), &(XMX[neig1*j]), neig*sizeof(PetscReal));
			CHKERRQ(ierr);
		}
	}
	else { /* Add XMX to T */
		for (j = 0; j < neig; j++){
			ierr=PetscMemcpy(&(TT[max_neig*j]), &(XMX[neig1*j]), neig*sizeof(PetscReal));
			CHKERRQ(ierr);
		}
	}
	
	dgmres->r += neig;
	r=dgmres->r;
	/*LU Factorize T with Lapack xgetrf routine */
	
	br = PetscBLASIntCast(r);
	bmax = PetscBLASIntCast(max_neig);
	if(!TTF){
		ierr = PetscMalloc(bmax*bmax*sizeof(PetscReal), &TTF); CHKERRQ(ierr);
	}
	ierr=PetscMemcpy(TTF, TT, bmax*r*sizeof(PetscReal)); CHKERRQ(ierr);
	if(!INVP){
		ierr=PetscMalloc(bmax*sizeof(PetscInt), &INVP); CHKERRQ(ierr);
	}
	LAPACKgetrf_(&r, &r, TTF, &bmax, INVP, &info);
	if (info) SETERRQ1(PETSC_ERR_LIB,"Error in LAPACK routine XGETRF INFO=%d",(int)info);
	
	/* Save X in U and MX in MU for the next cycles and increase the size of the invariant subspace */
	if(!UU){
		ierr=VecDuplicateVecs(VEC_VV(0), max_neig, &UU); CHKERRQ(ierr);
		ierr=VecDuplicateVecs(VEC_VV(0), max_neig, &MU); CHKERRQ(ierr);
	}
	for(j=0; j<neig; j++){
		ierr=VecCopy(XX[j], UU[r-neig+j]); CHKERRQ(ierr);
		ierr=VecCopy(MX[j], MU[r-neig+j]); CHKERRQ(ierr);
	}
	ierr=PetscLogEventEnd(KSP_DGMRESComputeDeflationData, ksp, 0,0,0); CHKERRQ(ierr);
	PetscFunctionReturn(0);
}

EXTERN_C_BEGIN
#undef __FUNCT__  
#define __FUNCT__ "KSPDGMRESImproveEig_DGMRES"
PetscErrorCode PETSCKSP_DLLEXPORT KSPDGMRESImproveEig_DGMRES(KSP ksp, PetscInt neig)
{
	KSP_DGMRES		*dgmres = (KSP_DGMRES*)ksp->data;
	PetscInt		j,i = 0, r_old, r = dgmres->r;
	PetscInt		neig1 = dgmres->neig + EIG_OFFSET; 
	PetscInt		bmax = dgmres->max_neig; 
	PetscInt		aug = r + neig; /* actual size of the augmented invariant basis */
	PetscInt		aug1 = bmax+neig1; /* maximum size of the augmented invariant basis */
	PetscBLASInt 	ldA; 		/* leading dimension of AUAU and AUU*/
	PetscBLASInt	N; 			/* size of AUAU */
	PetscReal		*Q; 		/*  orthogonal matrix of  (left) schur vectors */
	PetscReal		*Z; 		/*  orthogonal matrix of  (right) schur vectors */
	PetscReal 		*work;		/* working vector */
	PetscBLASInt 	lwork;		/* size of the working vector */
	PetscBLASInt 	info;		/* Output info from Lapack routines */
	PetscBLASInt 	*perm;		/* Permutation vector to sort eigenvalues */
	PetscReal		*wr, *wi, *beta, *modul; 	/* Real and imaginary part and modul of the eigenvalues of A*/
	PetscInt		ierr;
	
	PetscFunctionBegin;
	/* Block construction of the matrices AUU=(AU)'*U and (AU)'*AU*/
	if(!AUU){
		ierr=PetscMalloc(aug1*aug1*sizeof(PetscReal), &AUU); CHKERRQ(ierr);
		ierr=PetscMalloc(aug1*aug1*sizeof(PetscReal), &AUAU); CHKERRQ(ierr);
	}
	/* AUU = (AU)'*U = [(MU)'*U (MU)'*X; (MX)'*U (MX)'*X]
	 * Note that MU and MX have been computed previously either in ComputeDataDeflation() or down here in a previous call to this function */
	/* (MU)'*U size (r x r) -- store in the <r> first columns of AUU*/
	for(j=0; j < r; j++){
		ierr = VecMDot(UU[j], r, MU, &AUU[j*aug1]); CHKERRQ(ierr);
	}
	/* (MU)'*X size (r x neig) -- store in AUU from the column <r>*/
	for(j = 0; j < neig; j++){
		ierr = VecMDot(XX[j], r, MU, &AUU[(r+j)*aug1]); CHKERRQ(ierr);
	}
	/* (MX)'*U size (neig x r) -- store in the <r> first columns of AUU from the row <r>*/
	for(j = 0; j < r; j++){
		ierr = VecMDot(UU[j], neig, MX, &AUU[j*aug1+r]); CHKERRQ(ierr);
	}
	/* (MX)'*X size (neig neig) --  store in AUU from the column <r> and the row <r>*/
	for (j = 0; j < neig; j++){
		ierr = VecMDot(XX[j], neig, MX, &AUU[(r+j)*aug1 + r]); CHKERRQ(ierr);
	}
	
	/* AUAU = (AU)'*AU = [(MU)'*MU (MU)'*MX; (MX)'*MU (MX)'*MX] */
	/* (MU)'*MU size (r x r) -- store in the <r> first columns of AUAU*/
	for(j=0; j < r; j++){
		ierr = VecMDot(MU[j], r, MU, &AUAU[j*aug1]); CHKERRQ(ierr);
	}
	/* (MU)'*MX size (r x neig) -- store in AUAU from the column <r>*/
	for(j = 0; j < neig; j++){
		ierr = VecMDot(MX[j], r, MU, &AUAU[(r+j)*aug1]); CHKERRQ(ierr);
	}
	/* (MX)'*MU size (neig x r) -- store in the <r> first columns of AUAU from the row <r>*/
	for(j = 0; j < r; j++){
		ierr = VecMDot(MU[j], neig, MX, &AUAU[j*aug1+r]); CHKERRQ(ierr);
	}
	/* (MX)'*MX size (neig neig) --  store in AUAU from the column <r> and the row <r>*/
	for (j = 0; j < neig; j++){
		ierr = VecMDot(MX[j], neig, MX, &AUAU[(r+j)*aug1 + r]); CHKERRQ(ierr);
	}
	
	/* Computation of the eigenvectors */
	ldA = PetscBLASIntCast(aug1);
	N = PetscBLASIntCast(aug);
	lwork = 8 * N + 20; /* sizeof the working space */
	ierr = PetscMalloc(N*sizeof(PetscReal), &wr); CHKERRQ(ierr);
	ierr = PetscMalloc(N*sizeof(PetscReal), &wi); CHKERRQ(ierr);
	ierr = PetscMalloc(N*sizeof(PetscReal), &beta); CHKERRQ(ierr);
	ierr = PetscMalloc(N*sizeof(PetscReal), &modul); CHKERRQ(ierr);
	ierr = PetscMalloc(N*sizeof(PetscInt), &perm); CHKERRQ(ierr);
	ierr = PetscMalloc(N*N*sizeof(PetscReal), &Q); CHKERRQ(ierr);
	ierr = PetscMalloc(N*N*sizeof(PetscReal), &Z); CHKERRQ(ierr);
	ierr = PetscMalloc(lwork*sizeof(PetscReal), &work); CHKERRQ(ierr);
	LAPACKgges_("V", "V", "N", NULL, &N, AUAU, &ldA, AUU, &ldA, &i, wr, wi, beta, Q, &N, Z, &N, work, &lwork, NULL, &info);
	if( info){SETERRQ1(PETSC_ERR_LIB,"Error in LAPACK routine XGGES %d",(int)info);}
		
	for(i=0; i<N; i++){
		if( beta[i] !=0.0 ){
			wr[i] /=beta[i];
			wi[i] /=beta[i];
		}
	}
	/* sort the eigenvalues */
	for(i=0; i<N; i++)  {modul[i]=sqrt(wr[i]*wr[i]+wi[i]*wi[i]);}
	for (i=0; i<N; i++) { perm[i] = i;}
	ierr = PetscSortRealWithPermutation(N, modul, perm); CHKERRQ(ierr);
	/* Save the largest one */
	if(dgmres->lambdaN < modul[perm[N-1]] ) dgmres->lambdaN = modul[perm[N-1]];
	/* Allocate space to extract the first r schur vectors   */
	if(!SR2){
		ierr=PetscMalloc(aug1*bmax*sizeof(PetscReal), &SR2); CHKERRQ(ierr);
	} 
	/* count the number of extracted eigenvalues ( complex conjugates count as 2) */
	j=0;
	while( j < bmax ){
		i = perm[j];
		if(wi[i] == 0 ) j += 1; else j += 2;
	}
	if(j > bmax ) j = bmax - 1;
	r_old=r; /* previous size of r */
	/* extract the new <r> schur vectors */
	dgmres->r = r = j;
	for (j=0; j<r; j++){
		i = perm[j];
		ierr = PetscMemcpy(&SR2[j*aug1], &(Z[i*N]), N*sizeof(PetscReal)); CHKERRQ(ierr);
	}
	
	/* Multiply the Schur vectors SR2 by U (and X)  to get a new U 
	-- save it temporarily in MU */
	for (j = 0; j < r; j++){
		ierr = VecZeroEntries(MU[j]); CHKERRQ(ierr);
		ierr = VecMAXPY(MU[j], r_old, &SR2[j*aug1], UU); CHKERRQ(ierr);
		ierr = VecMAXPY(MU[j], neig, &SR2[j*aug1+r_old], XX); CHKERRQ(ierr);
	}
	/* Form T = U'*MU*U */
	for(j = 0; j < r; j++){
		ierr=VecCopy(MU[j], UU[j]); CHKERRQ(ierr);
		ierr = KSP_PCApplyBAorAB(ksp, UU[j], MU[j], VEC_TEMP_MATOP); CHKERRQ(ierr);
	}
	dgmres->matvecs += r;
	for(j = 0; j < r; j++){
		ierr = VecMDot(MU[j], r, UU, &TT[j*bmax]); CHKERRQ(ierr);		
	}
	/* Factorize T */
	ierr=PetscMemcpy(TTF, TT, bmax*r*sizeof(PetscReal)); CHKERRQ(ierr);
	LAPACKgetrf_(&r, &r, TTF, &bmax, INVP, &info);
	if (info) SETERRQ1(PETSC_ERR_LIB,"Error in LAPACK routine XGETRF INFO=%d",(int)info);
	/* Free Memory */
	PetscFree(wr); PetscFree(wi); PetscFree(beta); PetscFree(modul); 
	PetscFree(perm); PetscFree(Q); PetscFree(Z); PetscFree(work);
	

	PetscFunctionReturn(0);
}
EXTERN_C_END

EXTERN_C_BEGIN
#undef __FUNCT__  
#define __FUNCT__ "KSPDGMRESComputeSchurForm_DGMRES"
PetscErrorCode PETSCKSP_DLLEXPORT KSPDGMRESComputeSchurForm_DGMRES(KSP ksp, PetscInt *neig)
{
	KSP_DGMRES     	*dgmres = (KSP_DGMRES*)ksp->data;
	PetscErrorCode 	ierr; 
	PetscInt       	N = dgmres->max_k + 1, n=dgmres->it+1;
	PetscInt	bn, bN;
	PetscReal    	*A;
	PetscBLASInt 	ilo=1, ihi;
	PetscBLASInt 	ldA; 		/* leading dimension of A */
	PetscBLASInt	ldQ;		/* leading dimension of Q */
	PetscReal		*Q; 		/*  orthogonal matrix of  (left) schur vectors */
	PetscReal 		*work;		/* working vector */
	PetscBLASInt 	lwork;		/* size of the working vector */
	PetscBLASInt 	info;		/* Output info from Lapack routines */
	PetscBLASInt 	*perm;		/* Permutation vector to sort eigenvalues */
	PetscInt	i, j; 
	PetscReal	*wr, *wi, *modul; 	/* Real and imaginary part and modul of the eigenvalues of A*/
	
	PetscFunctionBegin;
	bn=PetscBLASIntCast(n);
	bN=PetscBLASIntCast(N);
	ihi = ldQ = bn;
	ldA=bN;
	lwork = PetscBLASIntCast(5*N);
	
#if defined(PETSC_USE_COMPLEX)
	SETERRQ(-1, "NO SUPPORT FOR COMPLEX VALUES AT THIS TIME");
	PetscFunctionReturn(-1);
#endif

	ierr = PetscMalloc(ldA*ldA*sizeof(PetscReal), &A); CHKERRQ(ierr);
	ierr = PetscMalloc(ldQ*n*sizeof(PetscReal), &Q); CHKERRQ(ierr);
	ierr = PetscMalloc(lwork*sizeof(PetscReal), &work); CHKERRQ(ierr);
	ierr = PetscMalloc(n*sizeof(PetscReal), &wr); CHKERRQ(ierr);
	ierr = PetscMalloc(n*sizeof(PetscReal), &wi); CHKERRQ(ierr);
	ierr = PetscMalloc(n*sizeof(PetscReal),&modul);CHKERRQ(ierr);
	ierr = PetscMalloc(n*sizeof(PetscInt),&perm);CHKERRQ(ierr);
	/* copy the Hessenberg matrix to work space */
	ierr = PetscMemcpy(A, dgmres->hes_origin, ldA*ldA*sizeof(PetscReal)); CHKERRQ(ierr);
	/* Compute eigenvalues with the Schur form */
	LAPACKhseqr_("S", "I", &bn, &ilo, &ihi, A, &ldA, wr, wi, Q, &ldQ, work, &lwork, &info);
	if (info) SETERRQ1(PETSC_ERR_LIB,"Error in LAPACK routine XHSEQR %d",(int)info);
	
	
	/* sort the eigenvalues */
	for(i=0; i<n; i++)  {modul[i]=sqrt(wr[i]*wr[i]+wi[i]*wi[i]);}
	for (i=0; i<n; i++) { perm[i] = i;}
	ierr = PetscSortRealWithPermutation(n, modul, perm); CHKERRQ(ierr);
	/* save the complex modulus of the largest eigenvalue in magnitude */
	if (dgmres->lambdaN < modul[perm[n-1]]){ dgmres->lambdaN=modul[perm[n-1]];}
	/* count the number of extracted eigenvalues (with complex conjugates) and extract the associated schur vectors*/
	j=0;
	while( j < dgmres->neig ){
		i = perm[j];
		ierr = PetscMemcpy(&SR[j*N], &(Q[i*ldQ]), n*sizeof(PetscReal)); CHKERRQ(ierr);
		j += 1; 
		if ( wi[i] != 0 ) {
			i = perm[j];
			ierr = PetscMemcpy(&SR[j*N], &(Q[i*ldQ]), n*sizeof(PetscReal));
			j += 1;
		}
	}
	*neig=j;
	ierr=PetscFree(A); 
	ierr=PetscFree(Q);
	ierr=PetscFree(wr);
	ierr=PetscFree(wi);
	ierr=PetscFree(work);
	ierr=PetscFree(modul); CHKERRQ(ierr);
	ierr=PetscFree(perm); CHKERRQ(ierr);
	PetscFunctionReturn(0);
}
EXTERN_C_END


EXTERN_C_BEGIN
#undef __FUNCT__  
#define __FUNCT__ "KSPDGMRESApplyDeflation_DGMRES"
PetscErrorCode PETSCKSP_DLLEXPORT KSPDGMRESApplyDeflation_DGMRES(KSP ksp, Vec x, Vec y)
{
	KSP_DGMRES     	*dgmres = (KSP_DGMRES*)ksp->data;
	PetscInt	i, r = dgmres->r;
	PetscErrorCode 	ierr; 
	PetscInt	info, nrhs = 1; 
	PetscReal	berr, ferr;
	PetscReal	alpha = 1.0;
	PetscInt	max_neig = dgmres->max_neig;
	PetscBLASInt	br = PetscBLASIntCast(r); 
	PetscBLASInt	bmax = PetscBLASIntCast(max_neig);
	PetscInt	lambda = dgmres->lambdaN;
	
	PetscFunctionBegin; 
	ierr=PetscLogEventBegin(KSP_DGMRESApplyDeflation, ksp, 0, 0, 0); CHKERRQ(ierr);
	if ( !r  ){
		ierr=VecCopy(x,y); CHKERRQ(ierr);
		PetscFunctionReturn(0);
	}
	/* Compute U'*x */
	if(!X1){
		ierr = PetscMalloc(bmax*sizeof(PetscReal), &X1); CHKERRQ(ierr);
		ierr = PetscMalloc(bmax*sizeof(PetscReal), &X2); CHKERRQ(ierr);
	}
	ierr = VecMDot(x, r, UU, X1); CHKERRQ(ierr);
	
	/* Solve T*X1=X2 for X1*/
	ierr = PetscMemcpy(X2, X1, br*sizeof(PetscReal)); CHKERRQ(ierr);
	LAPACKgetrs_("N", &br, &nrhs, TTF, &bmax, INVP, X1, &bmax, &info);
	if (info) SETERRQ1(PETSC_ERR_LIB,"Error in LAPACK routine XGETRS %d",(int)info);
	/* Iterative refinement -- is it really necessary ?? */
	if(!WORK){
		ierr=PetscMalloc(3*bmax*sizeof(PetscReal), &WORK); CHKERRQ(ierr);
		ierr=PetscMalloc(bmax*sizeof(PetscInt), &IWORK); CHKERRQ(ierr);
	}
	LAPACKgerfs_("N", &br, &nrhs, TT, &bmax, TTF, &bmax, INVP, X2, &bmax, 
				 X1, &bmax, &ferr, &berr, WORK, IWORK, &info);
	if (info) SETERRQ1(PETSC_ERR_LIB,"Error in LAPACK routine XGERFS %d",(int)info);
	
	for(i = 0; i < r; i++){
		X2[i] =  X1[i]/lambda - X2[i];
	}
    
	/* Compute X2=U*X2 */
	ierr=VecZeroEntries(y); CHKERRQ(ierr);
	ierr=VecMAXPY(y, r, X2, UU); CHKERRQ(ierr);
	
	ierr=VecAXPY(y, alpha, x); CHKERRQ(ierr);
	
	ierr=PetscLogEventEnd(KSP_DGMRESApplyDeflation, ksp, 0, 0, 0); CHKERRQ(ierr);
	PetscFunctionReturn(0);
}
EXTERN_C_END
/* end new DGMRES functions */

/*MC
     KSPDGMRES - Implements the deflated GMRES as defined by erhel, burrage and Pohl, NLAA 96
	GMRES Options Database Keys:
+   -ksp_gmres_restart <restart> - the number of Krylov directions to orthogonalize against
.   -ksp_gmres_haptol <tol> - sets the tolerance for "happy ending" (exact convergence)
.   -ksp_gmres_preallocate - preallocate all the Krylov search directions initially (otherwise groups of 
                             vectors are allocated as needed)
.   -ksp_gmres_classicalgramschmidt - use classical (unmodified) Gram-Schmidt to orthogonalize against the Krylov space (fast) (the default)
.   -ksp_gmres_modifiedgramschmidt - use modified Gram-Schmidt in the orthogonalization (more stable, but slower)
.   -ksp_gmres_cgs_refinement_type <never,ifneeded,always> - determine if iterative refinement is used to increase the 
                                   stability of the classical Gram-Schmidt  orthogonalization.
-   -ksp_gmres_krylov_monitor - plot the Krylov space generated
	DGMRES Options Database Keys:
	-ksp_dgmres_eigen <neig> - Number of smallest eigenvalues to extract at each restart 
	-ksp_dgmres_max_eigen <max_neig> - Maximum number of eigenvalues that can be extracted during the iterative process
	
   Level: beginner

   Notes: Left preconditioning is supported, but not right nor symmetric preconditioning. Complex arithmetic is not yet supported

   References:
     Restarted GMRES preconditioned by deflation,J. Computational and Applied Mathematics, 69(1996), 303-318.
	On the performance of various adaptive preconditioned GMRES strategies, 5(1998), 101-121.
	

.seealso:  KSPCreate(), KSPSetType(), KSPType (for list of available types), KSP, KSPFGMRES, KSPLGMRES,
           KSPGMRESSetRestart(), KSPGMRESSetHapTol(), KSPGMRESSetPreAllocateVectors(), KSPGMRESSetOrthogonalization(), KSPGMRESGetOrthogonalization(),
           KSPGMRESClassicalGramSchmidtOrthogonalization(), KSPGMRESModifiedGramSchmidtOrthogonalization(),
           KSPGMRESCGSRefinementType, KSPGMRESSetCGSRefinementType(), KSPGMRESGetCGSRefinementType(), KSPGMRESMonitorKrylov(), KSPSetPCSide()

M*/

EXTERN_C_BEGIN
#undef __FUNCT__  
#define __FUNCT__ "KSPCreate_DGMRES"
PetscErrorCode PETSCKSP_DLLEXPORT KSPCreate_DGMRES(KSP ksp)
{
  KSP_DGMRES      *dgmres;
  PetscErrorCode ierr;

  PetscFunctionBegin;
  ierr = PetscNewLog(ksp,KSP_DGMRES,&dgmres);CHKERRQ(ierr);
  ksp->data                              = (void*)dgmres;


  ksp->normtype                          = KSP_NORM_PRECONDITIONED;
  ksp->ops->buildsolution                = KSPBuildSolution_DGMRES;
  ksp->ops->setup                        = KSPSetUp_DGMRES;
  ksp->ops->solve                        = KSPSolve_DGMRES;
  ksp->ops->destroy                      = KSPDestroy_DGMRES;
  ksp->ops->view                         = KSPView_DGMRES;
  ksp->ops->setfromoptions               = KSPSetFromOptions_DGMRES;
  ksp->ops->computeextremesingularvalues = KSPComputeExtremeSingularValues_GMRES;
  ksp->ops->computeeigenvalues           = KSPComputeEigenvalues_GMRES;

  ierr = PetscObjectComposeFunctionDynamic((PetscObject)ksp,"KSPGMRESSetPreAllocateVectors_C",
                                    "KSPGMRESSetPreAllocateVectors_GMRES",
                                     KSPGMRESSetPreAllocateVectors_GMRES);CHKERRQ(ierr);
  ierr = PetscObjectComposeFunctionDynamic((PetscObject)ksp,"KSPGMRESSetOrthogonalization_C",
                                    "KSPGMRESSetOrthogonalization_GMRES",
                                     KSPGMRESSetOrthogonalization_GMRES);CHKERRQ(ierr);
ierr = PetscObjectComposeFunctionDynamic((PetscObject)ksp,"KSPGMRESSetRestart_C",
                                    "KSPGMRESSetRestart_GMRES",
                                     KSPGMRESSetRestart_GMRES);CHKERRQ(ierr);
  ierr = PetscObjectComposeFunctionDynamic((PetscObject)ksp,"KSPGMRESSetHapTol_C",
                                    "KSPGMRESSetHapTol_GMRES",
                                     KSPGMRESSetHapTol_GMRES);CHKERRQ(ierr);
  ierr = PetscObjectComposeFunctionDynamic((PetscObject)ksp,"KSPGMRESSetCGSRefinementType_C",
                                    "KSPGMRESSetCGSRefinementType_GMRES",
                                     KSPGMRESSetCGSRefinementType_GMRES);CHKERRQ(ierr);
 /* -- New functions defined in DGMRES -- */
  ierr=PetscObjectComposeFunctionDynamic((PetscObject)ksp, "KSPDGMRESSetEigen_C",
										  "KSPDGMRESSetEigen_DGMRES", 
											KSPDGMRESSetEigen_DGMRES); CHKERRQ(ierr);
  ierr=PetscObjectComposeFunctionDynamic((PetscObject)ksp, "KSPDGMRESSetMaxEigen_C",
										  "KSPDGMRESSetMaxEigen_DGMRES", 
											KSPDGMRESSetMaxEigen_DGMRES); CHKERRQ(ierr);
  ierr=PetscObjectComposeFunctionDynamic((PetscObject)ksp, "KSPDGMRESComputeSchurForm_C",
										  "KSPDGMRESComputeSchurForm_DGMRES", 
			KSPDGMRESComputeSchurForm_DGMRES); CHKERRQ(ierr);
  ierr=PetscObjectComposeFunctionDynamic((PetscObject)ksp, "KSPDGMRESComputeDeflationData_C",
										  "KSPDGMRESComputeDeflationData_DGMRES", 
			KSPDGMRESComputeDeflationData_DGMRES); CHKERRQ(ierr);
  ierr=PetscObjectComposeFunctionDynamic((PetscObject)ksp, "KSPDGMRESApplyDeflation_C",
										  "KSPDGMRESApplyDeflation_DGMRES", 
			KSPDGMRESApplyDeflation_DGMRES); CHKERRQ(ierr);
  ierr=PetscObjectComposeFunctionDynamic((PetscObject)ksp, "KSPDGMRESImproveEig_C", "KSPDGMRESImproveEig_DGMRES", KSPDGMRESImproveEig_DGMRES); CHKERRQ(ierr);
  
  ierr=PetscLogEventRegister("DGMRESComputeDefl", KSP_CLASSID, &KSP_DGMRESComputeDeflationData); CHKERRQ(ierr);
  ierr=PetscLogEventRegister("DGMRESApplyDefl", KSP_CLASSID, &KSP_DGMRESApplyDeflation); CHKERRQ(ierr);
  
  dgmres->haptol              = 1.0e-30;
  dgmres->q_preallocate       = 0;
  dgmres->delta_allocate      = GMRES_DELTA_DIRECTIONS;
  dgmres->orthog              = KSPGMRESClassicalGramSchmidtOrthogonalization;
  dgmres->nrs                 = 0;
  dgmres->sol_temp            = 0;
  dgmres->max_k               = GMRES_DEFAULT_MAXK;
  dgmres->Rsvd                = 0;
  dgmres->cgstype             = KSP_GMRES_CGS_REFINE_NEVER;
  dgmres->orthogwork          = 0;
  
  /* Default values for the deflation */
  dgmres->r					= 0;
  dgmres->neig				= DGMRES_DEFAULT_EIG;
  dgmres->max_neig			= GMRES_DEFAULT_MAXK-1;
  dgmres->lambdaN			= 0.0;
  dgmres->smv				= SMV;
  dgmres->force			= 0;
  dgmres->matvecs 			= 0;
  dgmres->improve			= 0;
  PetscFunctionReturn(0);
}
EXTERN_C_END
